/*
 * DiatomicInputStream.java
 *
 * Created on 11.1. 2004, 01:10
 */

package saiko.compression;

import java.io.ByteArrayOutputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;


/**
 * Class to provide decoding of data coded by DiatomicOutputStream
 *
 * @author  Dusan Saiko
 */
public class DiatomicInputStream extends FilterInputStream {
    
    byte buffer[]=null;     //data buffer into which data is expanded
    int  buffer_pos=-1;     //position in data buffer
    
    /** 
     * Constructor of new instance of BPEOutputStream
     *
     * @param aOutput (OutputStream) - output stream to write write the result into
     * @param aPassword (String) - password, minimum 9 characters
     *
     * @exception  IOException if an I/O error occurs and can not be written to the aOutput
    */
    public DiatomicInputStream(InputStream aInput) throws IOException
    {
        super(aInput);
        
        //read and compare the header
        byte header[]=new byte[DiatomicOutputStream.header.length()];
        in.read(header);
        if(!DiatomicOutputStream.header.equals(new String(header)))
            throw new IOException("Invalid data format - not Diatomic coding data stream");        
    }
    
   /**
     * Reads the next byte from this compressed output stream. 
     *
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized int read() throws IOException {
        
        //if buffer is empty or we have just read everything, reload the buffer
        if(buffer==null || buffer_pos>=buffer.length) {
            int length;
            length=in.read();
            // read next, if none then exit
            if(length<0) return length; //end of data

            Vector subst = new Vector();
            
            //read the substitutions
            for(int i=0; i<length; i++) {
                int pair1=in.read();
                int pair2=in.read();
                int substitute=in.read();
                subst.add(new DiatomicOutputStream.DiatomicSubstitution((byte)pair1, (byte)pair2,  (byte)substitute));
            }
            
            //length of the  compressed data chunk
            length=0;
            for(int i=0; i<3; i++) {
                length=length*256;
                length+=in.read();
            }
            
            byte data[]=new byte[length];
            in.read(data);

            //apply substitutions from the back
            ByteArrayOutputStream data2=new ByteArrayOutputStream(data.length);
            for(int i=subst.size()-1; i>=0; i--) {
                DiatomicOutputStream.DiatomicSubstitution current=(DiatomicOutputStream.DiatomicSubstitution) subst.get(i);
                for(int n=0; n<data.length; n++) {
                    if(current.substitute!=data[n]) {
                        data2.write(((int)(char)data[n])&0xFF);
                    } else {
                        data2.write(((int)(char)current.pair1)&0xFF);
                        data2.write(((int)(char)current.pair2)&0xFF);
                    }
                }
                data=data2.toByteArray();
                data2.reset();
            }
            
            buffer=data;
            buffer_pos=0;
        }
        
        //return next byte from the buffer
        return ((int)(char)buffer[buffer_pos++])&0xFF;
    }

    /**
     * reads the array of data
     * @param b - the buffer into which the data is read.
     * @param off - the start offset of the data.
     * @param len - the maximum number of bytes read. 
     * @return the total number of bytes read into the buffer, 
        or -1 if there is no more data because the end of the stream has been reached. 
     */
    public synchronized int read(byte b[], int off, int len)
    throws IOException
    {
        if ((off | len | (off + len) | (b.length - (off + len))) < 0) {
	    throw new IndexOutOfBoundsException();
	} else if (len == 0) {
	    return 0;
	}

	int n = 0;
	while (n < len) {
	    int n1 = read();
	    if (n1 < 0) break;
            b[off+n]=(byte)n1;
	    n ++;
	}
	return n==0 ? -1 : n;
    } 
} //class
