/*
 * Statistics.java
 *
 * Created on 4.1. 2004, 17:46
 */

package saiko.compression;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigInteger;

/**
 * class for holding the statistics informations
 *
 * @author  Dusan Saiko
 */
public class Statistics {
    
    byte[] data;    //data chunk
    int    count;   //number of times the data chunk occures in the text

    /** 
     * Constructor of new instance of Statistics
     * initialize counter for the data chunk to 0
     *
     * @param aData (byte[]) - data chunk
    */    
    public Statistics(byte[] aData)
    {
        data=aData;
        count=1;
    }
 
    /**
     * returns the number of occurance of the represented data
     */ 
    public int getCount()
    {
        return count;
    }
    
    /**
     * returns data for which this statistics is made
     */
    public byte[] getData() {
        return data;
    }
    
    /** 
     * Constructor of new instance of Statistics
     * initialize counter for the data chunk to 0
     *
     * @param aData (int[]) - data chunk, will be converted to byte[]
    */        
    public Statistics(int[] aData)
    {
        count=1;
        data=new byte[aData.length];
        for(int i=0; i<data.length; i++) {
            data[i]=(byte)aData[i];
        }
    }
    
    /**
     * returns numeric reprezentation of object content in the data[]
     */
    public BigInteger hash() {
        if(data==null) return BigInteger.ZERO;
        return new BigInteger(data);
    }
    
    public static void serialize(Statistics[] aStats, OutputStream aOutput) 
    throws IOException, IllegalArgumentException
    {
        OutputStream out=aOutput;
        int countSize=0;

        if(aStats==null) { throw new NullPointerException("Can not serialize null stats."); }
        
        int order=aStats[0].getData().length;
        if(order>31) throw new IllegalArgumentException("Maximum order of stats for serialization is 32 !");
        
        //get the largest count
        int max=0;
        for(int i=0; i<aStats.length; i++) {
            if(aStats[i].count>max) max=aStats[i].count;
        }
        //compute the number of bytes the count fits in
        int length=0;
        for(countSize=1; countSize<=8; countSize++) {
            length|=0xFF;
            length=length<<(countSize*8);
            if(length>max-1) break;
        }
        int v=order-1;
        v|=((countSize-1)<<5);
        out.write(v);
        
        //write number of stats pieces
        length=aStats.length-1;
        for(int i=0; i<order; i++) {
            out.write(length & 0xFF);
            length=length>>8;
        }
        
        //write each stat
        for(int i=0; i<aStats.length; i++) {
            for(int d=0; d<order; d++) {
                out.write(aStats[i].getData()[d]);
            }
            int count=aStats[i].getCount()-1;
            for(int x=0; x<countSize; x++) {
                out.write(count & 0xFF);
                count=count>>8;
            }
        }
        out.flush();
    }
    
    public static Statistics[] serialize(InputStream aIn) 
    throws IOException
    {
        int order=aIn.read();

        if(order==-1) return null;

        int countSize=((order&(0xFF<<5))>>5)+1;
        order=(order&(0xFF>>3))+1;
        
        //read the length of stats
        int length=0;
        for(int i=0; i<order; i++) {
            int n=aIn.read();
            length|=(n<<(8*i));
        }
        length+=1;
        
        Statistics stats[]=new Statistics[length];
        for(int i=0; i<stats.length; i++) {
            byte data[]=new byte[order];
            for(int d=0; d<order; d++) {
               data[d]=(byte)aIn.read();
            }
            stats[i]=new Statistics(data);
            int count=0;
            for(int x=0; x<countSize; x++) {
                count|=(aIn.read() << (8*x));
            }
            count+=1;
            stats[i].count=count;
        }        
        return stats;
    }
}
